package se.sek.any2zephyr.main;

import se.sek.any2zephyr.eventlogging.TransactionLog;
import se.sek.any2zephyr.zephyr.TestResult;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * Command line enabled start class for 'Any test result to Zephyr'
 * server start up a gateway listening for posted test run results from a
 * test run - and reporting the test results to Zephyr.
 *
 * Created by jordam on 2017-03-18.
 */
public class App {
    public static TransactionLog log = new TransactionLog();
    public static ProgramArgumentManager pam;
    private static boolean serverMode = false;
    public static boolean supressConnectionStatusChecking = false;

    /**
     * This is the main program executor
     *
     * @param args Runtime arguments
     */
    public static void main(String[] args){
        if(args.length == 0){
            System.out.println(helpText());
            System.exit(0);
        }

        pam =  new ProgramArgumentManager(args);
        log.log("Processing " + pam.numberOfArguments() + " runtime arguments.");
        setDebugModeIfApplicable();
        printHelpTextIfSwitchIsFound();

        Settings.attemptToReadFromFile();

        setAddress();
        setUserName();
        setPassword();

        saveSettingsIfWanted();

        cliUsage();
    }

    private static void printUnUsedArguments() {
        if(pam.unusedArguments().size() == 0)return;
        log.log("Unused arguments that could not be interpreted: '" + join("', '", pam.unusedArgumentsAsStrings().toArray(new String[0])) + "'.");
    }

    private static String createHelpSection(){
        return pam.createHelpSection()
                .addCliHint()
                .setMaxWidthInCharacters(78)
                .addArgument("Prints the help text.", null, false, "info", "man", "-help", "--help", "-h", "-?")
                .addArgument("Sets the URL to the Jira server.", "https://sek-jira.sek.se:8443", false, "jiraaddress")
                .addArgument("Identifies the issue to report test results to. A test called '" + Settings.testSummaryForCreatedTests + "' will be created if not found for the issue, if it's not a test issue.", "Automated regression test", false, "issue")
                .addArgument("Tells if the test should be set as passed or fail.", "pass|fail", false, "status")
                .addArgument("Adds a comment to the test execution in Jira/Zephyr", "\"Multi-word comment\"", false, "comment")
                .addArgument("Identifies the Jira project to find the issue in.", "SEK-Backlog", false, "project")
                .addArgument("Sets the Jira user name for interactions with Jira", "JenkinsJiraUser", false, "jirauser")
                .addArgument("Tells what password to use for the given Jira user. This could be stated in a JSON formatted any2zephyr.properties file in the executing folder.", "topSecret", false, "jirapassword")
                .addArgument("Saves the URL and login credentials to a JSON file for ease of use", null, false, "savesettings")
                .addArgument("More output from execution.", null, false, "debug")
                .toString();
    }

    private static void setDebugModeIfApplicable(){
        if(pam.hasArgument("debug")){
            Settings.debug = true;
            pam.use("debug");
        }
    }

    /**
     * Help text for command line.
     *
     * @return The text
     */
    private static String helpText(){
        return System.lineSeparator() +
                "Any test result to Zephyr server" + System.lineSeparator() +
                "===============================" + System.lineSeparator() +
                "This utility could both be used as s microservice or as a command line utility. Usage as a web server:" +
                "Usage example for server:" + System.lineSeparator() + System.lineSeparator() +
                "   java -jar any2zephyr.jar jiraaddress=http://jira.mycompany.com:80 username=taftestlinkuser password=secret [ port=2221 ][ timeout=120 ]" + System.lineSeparator() + System.lineSeparator() +
                "where port number is the AnyTest2Zephyr web server port to use." + System.lineSeparator() +
                "Default TCP port is 2221 - chosen not to collide with known services. The important part is that it should not be a port in use already." + System.lineSeparator() +
                System.lineSeparator() +
                "Usage example as a CLI utility:" + System.lineSeparator() +
                "   java -jar any2zephyr.jar project=MyJiraProjectName issue=MyIssueKey status=PASSED|FAILED comment=Yes!" + System.lineSeparator() + System.lineSeparator() +
                "If you want to understand how AnyTest2Zephyr server works, and should be applied, and what other switches exist, use the switch:" + System.lineSeparator() + System.lineSeparator() +
                "   java -jar any2zephyr.jar -help" + System.lineSeparator();
    }


    /**
     * Printing information text describing the workings of this program if the argument 'info' is found.
     */
    private static void printHelpTextIfSwitchIsFound(){
        if(pam.hasAnyOfTheArguments("info", "-help", "man", "--help", "-h")){
            pam.use("info", "-help", "man", "--help", "-h");
            System.out.println(System.lineSeparator() + "This server is used as a proxy between any test automation and Zephyr. It produces test run results to Zephyr. " + System.lineSeparator() + System.lineSeparator() +
                    "Starting the server" + System.lineSeparator() +
                    "--------------------------------------------------" + System.lineSeparator() +
                    "A few different command line parameters can be used at startup:" + System.lineSeparator() + System.lineSeparator() +
                    "Required parameters:" + System.lineSeparator() +
                    "     jiraaddress=http://jira.mycompany.com:80/" + System.lineSeparator() +
                    "                                              Make sure the full adress to the Jira" + System.lineSeparator() +
                    "                                              is there." + System.lineSeparator() + System.lineSeparator() +
                    "     jirauser=jirausername                    Used to interact with Jira, so" + System.lineSeparator() +
                    "                                              make sure you use a valid jiraUserName name for" + System.lineSeparator() +
                    "                                              Jira." + System.lineSeparator() + System.lineSeparator() +
                    "     jirapassword=yourpassword                Used to set the password to interact with Jira." + System.lineSeparator() + System.lineSeparator() +
                    "Optional parameters that change behaviour of this server:" + System.lineSeparator() +
                    "     mode=server|cli                          Set mode to server if you want to start" + System.lineSeparator() +
                    "                                              the REST server. Default is CLI usage." + System.lineSeparator() + System.lineSeparator() +
                    "     port=2221                                Port number is of your own choice for " + System.lineSeparator() +
                    "                                              activated. Make sure " + System.lineSeparator() +
                    "                                              the REST server, if server mode is" + System.lineSeparator() +
                    "                                              it is not the same as the Jira server uses." + System.lineSeparator() + System.lineSeparator() +
                    "     savesettings                             Saves Jira URL, user name, and password to a JSON formatted" + System.lineSeparator() +
                    "                                              file called 'any2zephyr.properties' in the program folder." + System.lineSeparator() + System.lineSeparator() +
                    "     config                                   Supresses Jira connection checking, for testing." + System.lineSeparator() + System.lineSeparator() +
                    "     project=MyProjectName                    Name of the Jira project for reporting test" + System.lineSeparator() +
                    "                                              results from CLI. Cannot containt spaces." + System.lineSeparator() + System.lineSeparator() +
                    "     issue=IssueKey                           For test result reporting from CLI only." + System.lineSeparator() +
                    "                                              The name of the issue key to report results" + System.lineSeparator() +
                    "                                              to. If the issue is a Zephyr test a new execution" + System.lineSeparator() +
                    "                                              will be created and the result reported to it. If" + System.lineSeparator() +
                    "                                              the issue is not a Zephyr test this utility will" + System.lineSeparator() +
                    "                                              make sure there is a Zephyr test called" + System.lineSeparator() +
                    "                                              '" + Settings.testSummaryForCreatedTests + "'" + System.lineSeparator() +
                    "                                              linked to the issue and report the result to this" + System.lineSeparator() +
                    "                                              test." + System.lineSeparator() + System.lineSeparator() +
                    "     status=pass|fail                         For test result reporting from CLI only." + System.lineSeparator() +
                    "                                              Test result to report." + System.lineSeparator() + System.lineSeparator() +
                    "     comment=My comment                       For test result reporting from CLI only." + System.lineSeparator() +
                    "                                              If used last in the CLI all trailing words" + System.lineSeparator() +
                    "                                              will be included in the comment. Strings" + System.lineSeparator() +
                    "                                              recognized as URI:s will be treated as links" + System.lineSeparator() +
                    "                                              in Zephyr." + System.lineSeparator() + System.lineSeparator() +
                    "All of these run time parameters are case insensitive and the order of them are irrelevant." + System.lineSeparator() + System.lineSeparator()+
                    "If the server mode is activated an output of interfaces will be printed. A HTML help page" + System.lineSeparator() +
                    "is available from running server. The link to this page is displayed when the server is starting up." + System.lineSeparator());
            System.exit(0);
        }
    }

    /**
     * Setting the address to the Jira installation.
     */
    private static void setAddress(){
        if(pam.hasArgument("jiraaddress")){
            log.log("Setting jiraAddress to " + pam.getValue("jiraaddress") + ".");
            Settings.jiraAddress = pam.use("jiraaddress");
        }
    }

    /**
     *  Sets the jiraUserName name used Jira. A username is used in
     *  combination with a API key.
     */
    private static void setUserName(){
        if(pam.hasArgument("jirauser")){
            log.log("Setting Jira jiraUserName name to " + pam.getValue("jirauser") + ".");
            Settings.jiraUserName = pam.use("jirauser");
        }
    }

    private static String join(String intermedium, String[] args){
        StringBuilder sb = new StringBuilder();
        for(int i = 0; i < args.length; i++){
            sb.append(args[0]);
            sb.append(intermedium);
        }
        String returnString = sb.toString();
        if(returnString.length() >= intermedium.length()){
            returnString = returnString.substring(0, returnString.length() - intermedium.length());
        }
        return returnString;
    }

    private static void cliUsage(){
        log.log("Assuming command line interface utility usage.");
        String project = pam.use("project");
        String issue = pam.use("issue");
        String status = pam.use("status");
        String comment = pam.use("comment");
        if(comment != null){
            List<String> usedArguments = new ArrayList<>();
            for(ProgramArgumentManager.Argument arg : pam.unusedArguments()){
                comment += " " + arg.argumentName;
                usedArguments.add(arg.argumentName);
            }
            for(String usedArg : usedArguments){
                pam.use(usedArg);
            }
        }
        printUnUsedArguments();
        TestResult testResult = new TestResult();
        if(project == null || issue == null || status == null){
            log.log("Need arguments for project, issue, and status.");
            return;
        }
        testResult.project = project;
        testResult.issue = issue;
        testResult.comment = comment;
        if(status.toLowerCase().equals("pass")) {
            testResult.status = TestResult.TestStatus.PASSED;
        } else if(status.toLowerCase().equals("fail")) {
            testResult.status = TestResult.TestStatus.FAILED;
        } else {
            testResult.status = TestResult.TestStatus.valueOf(status.toUpperCase());
        }
        testResult.report();
        log.log("Registered new test result from CLI usage: " + testResult.toString());
    }

    private static String replaceCharectersWithAsterisks(String instring){
        StringBuilder sb = new StringBuilder();
        for(Byte letter : instring.getBytes()){
            sb.append("*");
        }
        return sb.toString();
    }

    private static void setPassword() {
        if(pam.hasArgument("jirapassword")){
            log.log("Setting Jira password to '" + replaceCharectersWithAsterisks(pam.getValue("jirapassword")) + "'.");
            Settings.jiraPassword = pam.use("jirapassword");
        }
    }

    private static void saveSettingsIfWanted() {
        if(pam.hasArgument("savesettings")){
           Settings.attemptToSaveSettingsToFile();
           pam.use("savesettings");
           log.log("Saving current settings to 'any2zephyr.properties' file.");
        }
    }
}
